@@ -1,5 +1,6 @@
 import json
 import os
+import re
 from collections import OrderedDict
 
 import pandas as pd
@@ -9,12 +10,12 @@ from sim2bids.generate import subjects
 import sim2bids.templates.templates as temp
 
 # define naming conventions
-DEFAULT_TEMPLATE = '{}_desc-{}_{}.{}'
-COORD_TEMPLATE = 'desc-{}_{}.{}'
+DEFAULT_TEMPLATE = '{}_{}.{}'
+COORD_TEMPLATE = '{}.{}'
 
 # set to true if centres.txt (nodes and labels) are the same
 # for all files. In that case, store only one copy of the files
-# in the main 'coord' folder in the same scope as CHANGES.txt
+# (json, tsv) in the main 'coord' folder, in the global scope
 IGNORE_CENTRE = False
 
 # location of coord files (nodes and labels) in the scope of
@@ -22,7 +23,13 @@ IGNORE_CENTRE = False
 # sidecars, specifically `CoordsRows` and `CoordsColumns`
 COORDS = None
 
+# set true if `times` are the same for all files. In that case,
+# store only one copy of the files (json, tsv) in the main 'coord'
+# folder, in the global scope
+IGNORE_TIMES = False
+
 NETWORK = []
+TIMES_TO_SKIP = []
 
 
 def save(sub: dict, folders: list, ses: str = None, name: str = None) -> None:
@@ -72,7 +79,7 @@ def save(sub: dict, folders: list, ses: str = None, name: str = None) -> None:
     ----------
     sub (dict):
         Dictionary containing information of one file only. For example,
-        {'name': 'centres', 'fname': 'centres.txt', 'sid': 'sub-01', 'desc': 'default',
+        {'name': 'centres', 'fname': 'centres.txt', 'sid': '1', 'desc': 'default',
         'sep': '\t', 'path': PATH_TO_FILE, 'ext': 'txt'}
     folders (list):
         List of folders corresponding to whether input files have single- or
@@ -94,9 +101,12 @@ def save(sub: dict, folders: list, ses: str = None, name: str = None) -> None:
     """
     global IGNORE_CENTRE
 
+    if name == 'centres' and sub['name'] != 'centres':
+        name = 'coord'
+
     # check if centres should be ignored. If so, immediately break
     # the function. Otherwise, continue iteration.
-    if IGNORE_CENTRE and name == 'centres':
+    if IGNORE_CENTRE is True and name == 'centres' and os.path.exists(os.path.join(app.OUTPUT, 'coord', 'nodes.txt')):
         return
 
     # read file contents
@@ -111,7 +121,12 @@ def save(sub: dict, folders: list, ses: str = None, name: str = None) -> None:
             folder = folders[2]
 
         # get description for weights or distances
-        desc = temp.file_desc['weights'] if 'weights' in sub['name'] else temp.file_desc['distances']
+        if 'scthran' in sub['name'].lower():
+            desc = temp.file_desc['weights'].format('thresholded ')
+        elif 'scnot' in sub['name'].lower():
+            desc = temp.file_desc['weights'].format('')
+        else:
+            desc = temp.file_desc['distances']
 
         if 'content' in sub.keys():
             save_files(sub, folder, sub['content'], ftype='wd')
@@ -124,8 +139,10 @@ def save(sub: dict, folders: list, ses: str = None, name: str = None) -> None:
         save_centres(sub, file, ses, folders)
 
     # get folder location for spatial
-    elif name in ['spatial', 'fc']:
-        desc = temp.file_desc['spatial']
+    elif name in ['spatial', 'fc', 'map']:
+        fname = subjects.accepted(sub['name'], True)
+        desc = temp.file_desc['spatial_map'] if 'map' in fname \
+               else temp.file_desc[subjects.accepted(fname, return_accepted=True)].format(sub['sid'])
 
         if ses is None:
             folder = folders[2]
@@ -139,10 +156,13 @@ def save(sub: dict, folders: list, ses: str = None, name: str = None) -> None:
             save_files(sub, folder, file, type='default', ftype='spatial', desc=desc)
 
     # get folder location for time series
-    elif name in ['ts', 'bold']:
-        desc = temp.file_desc['ts'] if name == 'ts' else temp.file_desc['bold'].format(sub['sid'].replace('sub-', ''))
+    elif name in ['ts']:
+        desc = temp.file_desc['ts'].format(sub['sid']) if name == 'ts' else temp.file_desc['bold'].format(sub['sid'].replace('sub-', ''))
         save_files(sub, folders[-1], file, type='default', ftype='ts', desc=desc)
 
+    elif name == 'times':
+        traverse_times(sub, folders, ses)
+
     # get folder location for coordinates
     elif name == 'coord':
         # check nodes
@@ -170,15 +190,28 @@ def save(sub: dict, folders: list, ses: str = None, name: str = None) -> None:
                     else:
                         save_files(sub, folder, file, type='default', centres=True, desc=temp.centres['multi-unique'])
                 else:
-                    if IGNORE_CENTRE or not app.MULTI_INPUT:
-                        save_files(sub, folder, file, type='other', desc=temp.file_desc[sub['name']])
+                    accepted = subjects.get_name(sub['name'], True)
+                    if IGNORE_CENTRE and 'ses' in folder:
+                        save_files(sub, folder, file, type='default', desc=temp.file_desc[accepted])
+                    elif IGNORE_CENTRE and 'ses' not in folder and app.SESSIONS:
+                        save_files(sub, folder, file, type='other', desc=temp.file_desc[accepted])
                     else:
-                        save_files(sub, folder, file, type='default', desc=temp.file_desc[sub['name']])
+                        accepted = subjects.get_name(sub['name'])
+
+                        if IGNORE_CENTRE and 'sub' in folder:
+                            type_ = 'default'
+                        else:
+                            type_ = 'coord'
+
+                        save_files(sub, folder, file, type=type_, desc=temp.file_desc[accepted])
 
 
 def save_centres(sub, file, ses, folders, centre_name='centres'):
     global IGNORE_CENTRE
 
+    if sub['name'] != 'centres':
+        save(sub, folders, ses, 'coord')
+
     # check if all centres are of the same content
     if check_centres(name=centre_name):
         # ignore preceding centres files
@@ -194,18 +227,21 @@ def save_centres(sub, file, ses, folders, centre_name='centres'):
         desc = temp.centres['multi-unique'] if app.MULTI_INPUT else temp.centres['single']
 
         # set appropriate output path depending on session and subject types
-        if ses is not None:
-            folder = folders[4]
-        elif app.MULTI_INPUT and ses is None:
-            folder = folders[3]
+        if len(folders) > 1:
+            if ses is not None:
+                folder = folders[4]
+            elif app.MULTI_INPUT and ses is None:
+                folder = folders[3]
+            else:
+                folder = os.path.join(app.OUTPUT, 'coord')
         else:
-            folder = os.path.join(app.OUTPUT, 'coord')
+            folder = folders[0]
 
         # save conversion results
         if IGNORE_CENTRE:
-            save_files(sub, folder, file, type='other', centres=True, desc=desc)
-        else:
             save_files(sub, folder, file, type='default', centres=True, desc=desc)
+        else:
+            save_files(sub, folder, file, type='other', centres=True, desc=desc)
 
 
 def save_h5(sub, folders, ses=None):
@@ -224,7 +260,7 @@ def save_files(sub: dict, folder: str, content, type: str = 'default', centres:
     ----------
     sub (dict):
         Dictionary containing information of one file only. For example,
-        {'name': 'centres', 'fname': 'centres.txt', 'sid': 'sub-01', 'desc': 'default',
+        {'name': 'centres', 'fname': 'centres.txt', 'sid': '1', 'desc': 'default',
         'sep': '\t', 'path': PATH_TO_FILE, 'ext': 'txt'}
     folder (str):
         Folder where to store output conversions for the specific file. For example,
@@ -257,11 +293,14 @@ def save_files(sub: dict, folder: str, content, type: str = 'default', centres:
     """
     global COORDS
 
+    name = sub['name'].replace('.txt', '')
+
     if type == 'default':
-        json_file = os.path.join(folder, DEFAULT_TEMPLATE.format(sub['sid'], sub['desc'], sub['name'], 'json'))
+        json_file = os.path.join(folder, DEFAULT_TEMPLATE.format(sub['sid'], name, 'json'))
         tsv_file = json_file.replace('json', 'tsv')
+
     else:
-        json_file = os.path.join(folder, COORD_TEMPLATE.format(sub['desc'], sub['name'], 'json'))
+        json_file = os.path.join(folder, COORD_TEMPLATE.format(name, 'json'))
         tsv_file = json_file.replace('json', 'tsv')
 
     # Save 'centres.txt' as 'nodes.txt' and 'labels.txt'. This will require breaking the
@@ -293,14 +332,12 @@ def save_files(sub: dict, folder: str, content, type: str = 'default', centres:
         to_tsv(nodes.replace('json', 'tsv'), content.iloc[:, 1:])
     else:
         if ftype == 'coord':
-            if sub['name'] == 'bold_times':
-                desc = temp.file_desc['bold_times'].format(sub['sid'])
-            to_json(json_file.lower(), shape=content.shape, key='coord', desc=desc)
-            to_tsv(tsv_file.lower(), content)
+            to_json(json_file, shape=content.shape, key='coord', desc=desc)
+            to_tsv(tsv_file, content)
         else:
             # otherwise, save files as usual
-            to_json(json_file.lower(), shape=content.shape, key=ftype, desc=desc)
-            to_tsv(tsv_file.lower(), content)
+            to_json(json_file, shape=content.shape, key=ftype, desc=desc)
+            to_tsv(tsv_file, content)
 
 
 def check_centres(name='centres'):
@@ -319,29 +356,35 @@ def check_centres(name='centres'):
     # get all centres files
     centres = get_specific(name)
 
-    # get the first element
-    file = open_file(centres[0], subjects.find_separator(centres[0]))
+    if len(centres) == 1:
+        return True
 
-    # define set literal
-    same = {}
+    if centres:
+        # get the first element
+        file = open_file(centres[0], subjects.find_separator(centres[0]))
 
-    # iterate over centres
-    for centre in centres[1:]:
-        # append to the set literal whether the contents of the first element
-        # are the same with the rest
-        same.update(file == open_file(centre, subjects.find_separator(centre)))
+        # define set literal
+        same = {}
 
-    # check if set literal contains only one element
-    if len(same) == 1:
-        if bool(same):
-            # if files are the same, return True
-            return True
-        # False otherwise
+        # iterate over centres
+        for centre in centres[1:]:
+            # append to the set literal whether the contents of the first element
+            # are the same with the rest
+            same.update(file == open_file(centre, subjects.find_separator(centre)))
+
+        # check if set literal contains only one element
+        if len(same) == 1:
+            if bool(same):
+                # if files are the same, return True
+                return True
+            # False otherwise
+            return False
         return False
-    return False
+
+    return None
 
 
-def get_specific(filetype: str) -> list:
+def get_specific(filetype: str, constraint: str = None) -> list:
     """
     Get all files that correspond to the filetype. For example,
     if filetype is equal to "areas", this function will return
@@ -362,6 +405,7 @@ def get_specific(filetype: str) -> list:
     Returns
     -------
         Returns a list of all files that end with the specified filetype.
+        :param constraint:
     """
 
     # create emtpy list to store appropriate files
@@ -371,8 +415,11 @@ def get_specific(filetype: str) -> list:
     for file in app.ALL_FILES:
         # check if the keyword is present
         if filetype in file:
-            # if yes, append the value
-            content.append(file)
+            if constraint is not None and constraint in file:
+                content.append(file)
+            if constraint is None:
+                # if yes, append the value
+                content.append(file)
 
     # return the list of newly-found files
     return content
@@ -407,6 +454,53 @@ def open_file(path: str, sep: str):
         pass
 
 
+def traverse_times(sub, folders, ses):
+    global TIMES_TO_SKIP
+
+    # get description
+    desc = temp.file_desc['times'] if 'bold' not in sub['name'] else temp.file_desc['bold_times']
+
+    # check if times are similar
+    for times in app.TIMES:
+        for rhythm in ['alpha', 'delta', 'beta', 'gamma', 'theta']:
+            # check if the file needs to be skipped
+            if f'{rhythm}_{times}' in TIMES_TO_SKIP:
+                return
+
+            results = get_specific(times, rhythm)
+
+            if len(results) > 0:
+                open_df = lambda x: pd.read_csv(x, header=None, sep='\t')
+                first = open_df(results[0])
+
+                if len(results) > 1:
+                    for result in results[1:]:
+                        if first.equals(open_df(result)):
+                            similar = True
+                            TIMES_TO_SKIP.append(f'{rhythm}_{times}')
+
+                            # save in global folder
+                            sub['name'] = f'{rhythm}-{times}'
+                            save_files(sub, os.path.join(app.OUTPUT, 'coord'), first, 'coord', desc=desc)
+                        else:
+                            # set appropriate output path depending on session and subject types
+                            if ses is not None:
+                                folder = folders[4]
+                            elif app.MULTI_INPUT and ses is None:
+                                folder = folders[3]
+                            else:
+                                folder = os.path.join(app.OUTPUT, 'coord')
+
+                            save_files(sub, folder, first, 'default', desc=desc)
+                            save_files(sub, folder, open_df(result), 'default', desc=desc)
+                else:
+                    similar = True
+                    TIMES_TO_SKIP.append(f'{rhythm}_{times}')
+                    # save in global folder
+                    sub['name'] = f'{rhythm}-{times.replace(".txt", "")}'
+                    save_files(sub, os.path.join(app.OUTPUT, 'coord'), first, 'coord', desc=desc)
+
+
 def open_text(path, sep):
     """
 
@@ -425,7 +519,7 @@ def open_text(path, sep):
         return pd.read_csv(path, header=None, index_col=None)
 
     try:
-        f = pd.read_csv(path, sep=sep, header=None, index_col=False)
+        f = pd.read_csv(path, sep=sep, header=None, index_col=None)
     except pd.errors.EmptyDataError:
         return ''
     except ValueError:
@@ -486,35 +580,91 @@ def to_json(path, shape, desc, key, **kwargs):
     -------
 
     """
+    global NETWORK
 
-    inp = temp.required
-    out = OrderedDict({x: '' for x in inp})
-
-    params = {'ModelEq': f'../eq/desc-{app.DESC}_eq.xml',
-              'ModelParam': f'../param/desc-{app.DESC}_param.xml',
-              'SourceCode': f'../code/desc-{app.DESC}_code.py',
-              'SoftwareVersion': app.SoftwareVersion,
-              'SoftwareRepository': app.SoftwareRepository,
-              'SourceCodeVersion': app.SoftwareVersion,
-              'SoftwareName': app.SoftwareName,
-              'Network': NETWORK
-              }
+    if key not in ['param', 'eq', 'code']:
+        out = OrderedDict({x: '' for x in temp.required})
+    else:
+        out = OrderedDict()
 
     if key != 'wd':
         struct = temp.struct[key]
         out.update({x: '' for x in struct['required']})
 
-    for k in ['ModelEq', 'ModelParam', 'SourceCode', 'SoftwareVersion', 'SourceCodeVersion',
-              'SoftwareRepository', 'SoftwareName', 'Network']:
-        if k in out.keys():
-            out[k] = params[k]
+    # ===========================================================
+    #                     TAKE CARE OF COORDS
+    # ===========================================================
+    coord = None
 
-    if 'Units' in out.keys() and key != 'coord':
-        out['Units'] = 'ms'
+    if 'CoordsRows' in out.keys() or key in ['wd', 'ts', 'spatial']:
+        if app.SESSIONS and IGNORE_CENTRE:
+            coord = ['../../../coord/nodes.json', '../../../coord/labels.json']
+        elif app.SESSIONS and not IGNORE_CENTRE:
+            coord = ['../coord/nodes.json', '../coord/labels.json']
+        elif not app.SESSIONS and IGNORE_CENTRE:
+            coord = ['../../coord/nodes.json', '../../coord/labels.json']
+        else:
+            coord = ['../coord/nodes.json', '../coord/labels.json']
 
-    # point coord files to nodes/labels TSV files instead
-    coord = COORDS if key != 'coord' else [COORDS[0].replace('json', 'tsv'), COORDS[1].replace('json', 'tsv')] \
-        if COORDS is not None else COORDS
+    if key != 'wd':
+
+        # ===========================================================
+        #                     TAKE CARE OF EQUATIONS
+        # ===========================================================
+
+        if 'ModelEq' in out.keys() or 'ModelEq' in temp.struct[key]['recommend']:
+            if 'param' in path or 'code' in path:
+                out['ModelEq'] = '../eq/eq.xml'
+            else:
+                if app.SESSIONS:
+                    out['ModelEq'] = '../../../eq/eq.xml'
+                else:
+                    out['ModelEq'] = '../../eq/eq.xml'
+
+        # ===========================================================
+        #                     TAKE CARE OF PARAMETERS
+        # ===========================================================
+
+        if 'ModelParam' in out.keys() or 'ModelParam' in temp.struct[key]['recommend']:
+            if 'eq' in path or 'code' in path:
+                out['ModelParam'] = '../param/param.xml'
+            else:
+                if app.SESSIONS:
+                    out['ModelParam'] = '../../../param/param.xml'
+                else:
+                    out['ModelParam'] = '../../param/param.xml'
+
+        params = {
+            'SourceCode': f'/tvb-framework-{app.SoftwareVersion}' if app.SoftwareCode == 'MISSING' else app.SoftwareCode,
+            'SoftwareVersion': app.SoftwareVersion if app.SoftwareVersion else None,
+            'SoftwareRepository': app.SoftwareRepository if app.SoftwareRepository else None,
+            'SourceCodeVersion': app.SoftwareVersion if app.SoftwareVersion else None,
+            'SoftwareName': app.SoftwareName if app.SoftwareName else None,
+            'Network': NETWORK if NETWORK else None
+        }
+
+        if 'eq' not in path or 'param' not in path or 'code' not in path:
+            sub = re.findall(r'sub-[0-9]+', path)
+            network = None
+
+            if sub:
+                network = [f'../net/{sub[0]}_{app.DESC}_weights.json', f'../net/{sub[0]}_{app.DESC}_distances.json']
+
+            if network:
+                params['Network'] = network
+
+        if kwargs:
+            params += kwargs
+
+        for k in params.keys():
+            if k in temp.struct[key]['required'] or k in temp.struct[key]['recommend']:
+                out[k] = params[k]
+
+    if 'Units' in out.keys():
+        if key == 'wd' or key == 'coord':
+            out['Units'] = ''
+        else:
+            out['Units'] = 'ms'
 
     with open(path, 'w') as file:
-        json.dump(temp.populate_dict(out, shape=shape, desc=desc, coords=coord, **kwargs), file)
+        json.dump(temp.populate_dict(out, shape=shape, desc=desc, coords=coord), file)
